from .data_types import *
import os
import logging

all_classes = [
    StateArray,
    SwitchStates,
    CapacitorStates,
    RegulatorStates,
    CostArray,
    RealCostFunctions,
    ReactiveCostFunctions,
    RealWholesalePrices,
    ReactiveWholesalePrices,
    OperationalCosts,
    MeasurementArray,
    VoltagesMagnitude,
    VoltagesAngle,
    VoltagesReal,
    VoltagesImaginary,
    CurrentsMagnitude,
    CurrentsAngle,
    CurrentsReal,
    CurrentsImaginary,
    PowersMagnitude,
    PowersAngle,
    PowersReal,
    PowersImaginary,
    PowersImaginary,
    SolarIrradiances,
    Temperatures,
    WindSpeeds,
    StatesOfCharge,
    Topology,
    AdmittanceSparse,
    AdmittanceMatrix,
    Injection,
    Command,
    CommandList,
    VVControl,
    VWControl,
    InverterControl,
    InverterControlList,
]

if not os.path.exists("schemas"):
    os.makedirs("schemas")

for klass in all_classes:
    logging.info(klass)
    json_data = klass.schema_json(indent=4)
    with open(os.path.join("schemas", klass.__name__ + ".json"), "w") as fp:
        fp.write(json_data)



"""Wiring Diagram utilities

Wiring diagrams can be hard to manage in their final list based form.
Some utilities plot, and future additions include nested wiring
diagram composition and a programmatic interface.
"""

from . import system_configuration
from .system_configuration import WiringDiagram

def get_graph(wiring_diagram: WiringDiagram):
    "Get networkx graph representation of wiring_diagram"
    import networkx as nx

    g = nx.MultiDiGraph()
    for c in wiring_diagram.components:
        g.add_node(c.name, type=c.type, parameters=c.parameters)
    for l in wiring_diagram.links:
        g.add_edge(l.source, l.target,
                   source_port=l.source_port,
                   target_port=l.target_port)
    return g


def plot_graph_matplotlib(wiring_diagram: WiringDiagram):
    import matplotlib.pyplot as plt
    import networkx as nx

    g = get_graph(wiring_diagram)
    pos = nx.spring_layout(g)
    nx.draw(
        g, pos, edge_color='black', width=1, linewidths=1,
        node_size=500, alpha=0.9,
        labels={node: node for node in g.nodes()}
    )

    edge_map = {
        (l.source, l.target): f"{l.source_port} -> {l.target_port}"
        for l in wiring_diagram.links
    }

    nx.draw_networkx_edge_labels(
        g, pos, edge_map,
        font_color='red'
    )
    plt.axis('off')
    plt.show()


def get_graph_renderer(G):
    import networkx as nx
    from bokeh.plotting import from_networkx
    from bokeh.models import (Circle, EdgesOnly, MultiLine)
    from bokeh.palettes import Spectral4

    graph_renderer = from_networkx(G, nx.spectral_layout, scale=1, center=(0,0))

    graph_renderer.node_renderer.glyph = Circle(size=15, fill_color=Spectral4[0])
    graph_renderer.node_renderer.selection_glyph = Circle(size=15, fill_color=Spectral4[2])
    graph_renderer.node_renderer.hover_glyph = Circle(size=15, fill_color=Spectral4[1])

    graph_renderer.edge_renderer.glyph = MultiLine(line_color="#CCCCCC", line_alpha=0.8, line_width=5)
    graph_renderer.edge_renderer.selection_glyph = MultiLine(line_color=Spectral4[2], line_width=5)
    graph_renderer.edge_renderer.hover_glyph = MultiLine(line_color=Spectral4[1], line_width=5)

    #graph_renderer.selection_policy = NodesAndLinkedEdges()
    graph_renderer.inspection_policy = EdgesOnly()
    return graph_renderer


def plot_graph_bokeh(wiring_diagram: WiringDiagram):
    from bokeh.models import (BoxSelectTool, HoverTool,
                              Plot, Range1d, TapTool,
                              PreText, CustomJS)
    from bokeh.layouts import layout
    from bokeh.io import output_file, show

    G = get_graph(wiring_diagram)
    graph_renderer = get_graph_renderer(G)
    source = graph_renderer.node_renderer.data_source

    plot = Plot(width=800, height=800,
                x_range=Range1d(-1.1,1.1), y_range=Range1d(-1.1,1.1))
    plot.title.text = f"{wiring_diagram.name} cosimulation"
    plot.renderers.append(graph_renderer)

    node_text = PreText(
        text='No node seleted', width=500, height=500,
    )
    edge_hover_tool = HoverTool(tooltips=[
        ("source port", "@source_port"),
        ("target port", "@target_port")
    ])
    tap_tool = TapTool()
    tap_tool.callback = CustomJS(
        args={"nodes": source, "node_text": node_text}, code="""
        var ind = cb_data.source.selected.indices[0];
        var name = nodes.data['index'][ind];
        var type = nodes.data['type'][ind];
        var parameters = nodes.data['parameters'][ind];
        var s = "Name: " + name + ". ";
        s += "Type: " + type + ". " ;
        s += "Parameters: " + JSON.stringify(parameters, null, 2);

        node_text.text = s;
        node_text.width = 500;
        style="white-space: pre-wrap;"
    """)

    plot.add_tools(edge_hover_tool, tap_tool, BoxSelectTool())


    layout_ = layout([[plot,node_text]])
    show(layout_)



import time

import click
import helics as h
from .broker_utils import pprint_time_data, get_time_data


class PausingBroker:
    def __init__(self, n):
        self.initstring = f"-f {n} --name=mainbroker"

    def run(self):
        self.broker = h.helicsCreateBroker("zmq", "", self.initstring)
        print("Setting time barrier to 0.0")
        h.helicsBrokerSetTimeBarrier(self.broker, 0.0)
        t = 0.0

        time.sleep(3)
        name_2_timedata = {}
        while h.helicsBrokerIsConnected(self.broker) is True:
            for time_data in get_time_data(self.broker):
                if ((time_data.name not in name_2_timedata) or
                    (name_2_timedata[time_data.name] != time_data)):
                    name_2_timedata[time_data.name] = time_data
                    pprint_time_data(time_data)

            new_t = click.prompt('Enter next time:', type=float, default=t)
            if new_t != t:
                t = new_t
                print(f"Setting time barrier to {t}")
                h.helicsBrokerSetTimeBarrier(self.broker, t)



import click
import json
import os
import subprocess
import time

from oedisi.componentframework.basic_component import (
    basic_component,
    ComponentDescription
)
from oedisi.componentframework.mock_component import MockComponent
from oedisi.componentframework.system_configuration import (
    RunnerConfig,
    generate_runner_config,
    WiringDiagram,
    Component,
)
from .pausing_broker import PausingBroker
from .testing_broker import TestingBroker

@click.group()
def cli():
    pass


def bad_type_checker(type, x):
    "Does not check types"
    return True


def get_basic_component(filename):
    # before, the runner would use the directory given _in_ the component description
    # which may be inaccurate
    comp_desc = ComponentDescription.parse_file(filename)
    comp_desc.directory = os.path.dirname(filename)  
    return basic_component(comp_desc, bad_type_checker)


@cli.command()
@click.option("--target-directory", default="build", type=click.Path(),
              help="Target directory to put the system in")
@click.option("--system", default="system.json", type=click.Path(),
              help="Wiring diagram json to build")
@click.option("--component-dict", default="components.json", type=click.Path(),
              help="path to JSON Dictionary of component folders")
def build(target_directory, system, component_dict):
    """Build to the simulation folder

    Examples::

        oedisi build

        oedisi build --component-dict components.json --system scenario.json

    \f
    Parameters
    ----------

    target_directory : str (default="build")
        build path
    system : str (default="system.json")
        path to wiring diagram json
    component_dict : str (default="components.json")
        path to JSON dictionary of component folders
    """
    click.echo(f"Loading the components defined in {component_dict}")
    with open(component_dict, 'r') as f:
        component_dict_of_files = json.load(f)
        component_types = {
            name: get_basic_component(component_file)
            for name, component_file in component_dict_of_files.items()
        }

    click.echo(f"Loading system json {system}")
    wiring_diagram = WiringDiagram.parse_file(system)

    click.echo(f"Building system in {target_directory}")
    runner_config = generate_runner_config(
        wiring_diagram, component_types, target_directory=target_directory
    )
    with open(f"{target_directory}/system_runner.json", "w") as f:
        f.write(runner_config.json(indent=2))


@cli.command()
@click.option("--runner", default="build/system_runner.json", type=click.Path(),
              help="Location of helics run json. Usually build/system_runner.json")
def run(runner):
    """Calls out to helics run command

    Examples::

        oedisi run
    """
    subprocess.run(["helics", "run", f"--path={runner}"])


@cli.command()
@click.option("--runner", default="build/system_runner.json", type=click.Path(),
              help="Location of helics run json. Usually build/system_runner.json")
def run_with_pause(runner):
    """Helics broker is run in the foreground, and we allow user input
    to block time.

    Examples::

        oedisi run-with-pause


        Enter next time: [0.0]: 1.0
        Setting time barrier to 1.0

            Name         : comp_abc
            Granted Time : 0.0
            Send Time    : 0.0


            Name         : comp_xyz
            Granted Time : 0.0
            Send Time    : 0.0

    """
    new_system, new_path, _ = remove_from_json(runner, "broker")
    background_runner = subprocess.Popen(["helics", "run", f"--path={new_path}"])
    broker = PausingBroker(len(new_system.federates))
    broker.run()
    background_runner.wait()


@cli.command()
@click.option("--target-directory", default="build", type=click.Path(),
              help="Target directory to put the system in")
@click.option("--component-desc", type=click.Path(),
              help="Path to component description")
@click.option("--parameters", default=None,
              help="Path to parameters JSON (default is parameters={})")
def test_description(target_directory, component_desc, parameters):
    """Test component intialization from component description

    Examples::

        oedisi test-description --component-desc component/component_definition.json --parameters inputs.json

        Initialized broker
        Waiting for initialization
        Testing dynamic input names
        ✓
        Testing dynamic output names
        ✓


    \f
    Parameters
    ----------

    target_directory : str
        build location

    component_desc : str
        filepath to component_description.json to test

    parameters : str
        filepath to parameters json (default is parameters={})

    Process
    -------

    Get inputs and outputs from component_desc
    Create a wiring diagram for component
        and a "do-nothing" component with parameters for the corresponding
        inputs and outputs (basically recorder federate?)
    Create and run system with wiring diagram
    """

    comp_desc = ComponentDescription.parse_file(component_desc)
    comp_desc.directory = os.path.dirname(component_desc)

    if parameters is None:
        parameters = {}
    else:
        with open(parameters) as f:
            parameters = json.load(f)

    inputs = list(map(lambda x: x.port_name, comp_desc.dynamic_inputs))
    outputs = list(map(lambda x: x.port_name, comp_desc.dynamic_outputs))
    w = WiringDiagram.empty()
    component = Component(name="component", type="UserComponent",
                          parameters=parameters)
    tester = Component(name="tester", type="MockComponent", parameters={
        "inputs": outputs,
        "outputs": {x.port_name: x.type for x in comp_desc.dynamic_inputs}
    })
    w.add_component(component)
    w.add_component(tester)
    for i in outputs:
        w.add_link(component.port(i).connect(tester.port(i)))

    for i in inputs:
        w.add_link(tester.port(i).connect(component.port(i)))

    component_types = {
        "MockComponent": MockComponent,
        "UserComponent": basic_component(
            comp_desc,
            bad_type_checker
        )
    }
    runner_config = generate_runner_config(w, component_types, target_directory=target_directory)
    runner_config, _ = remove_from_runner_config(runner_config, "broker")
    with open(f"{target_directory}/system_runner.json", "w") as f:
        f.write(runner_config.json())

    background_runner = subprocess.Popen(
        ["helics", "run", f"--path={target_directory}/system_runner.json"]
    )
    broker = TestingBroker(len(runner_config.federates))
    federate_inputs, federate_outputs = broker.run()
    background_runner.kill()
    print("Testing dynamic input names")
    assert sorted(list(map(
        lambda x: x.port_name, comp_desc.dynamic_inputs
    ))) == sorted(list(map(
        lambda x: x.split("/")[1], federate_inputs["component"]
    )))
    print("✓")
    print("Testing dynamic output names")
    assert sorted(list(map(
        lambda x: "component/" + x.port_name, comp_desc.dynamic_outputs
    ))) == sorted(federate_outputs["component"])
    print("✓")



def remove_from_runner_config(runner_config, element):
    "Remove federate from configuration"
    within_feds = [fed for fed in runner_config.federates
                        if fed.name != element]
    without_feds = [fed for fed in runner_config.federates
                        if fed.name == element]
    new_config = RunnerConfig(
        name = runner_config.name,
        federates = within_feds
    )
    return new_config, without_feds


def remove_from_json(system_json, element):
    "Remove federate from configuration and resave with revised.json"
    with open(system_json, "r") as f:
        runner_config = RunnerConfig.parse_obj(json.load(f))
        new_config, without_feds = remove_from_runner_config(
            runner_config,
            element
        )

        new_path = system_json + "revised.json"
        click.echo(f"Saving new json to {new_path}")
        with open(new_path, "w") as f:
            f.write(new_config.json())
        return new_config, new_path, without_feds


@cli.command()
@click.option("--runner", default="build/system_runner.json", type=click.Path(),
              help="Location of helics run json. Usually build/system_runner.json")
@click.option("--foreground", type=str, help="Name of component to run in background")
def debug_component(runner, foreground):
    """
    Run system runner json with one component in the JSON

    We remove one component from system_runner.json
    and then call helics run in the background with our new json.
    and then run our debugging component in standard in / standard out.

    \f
    Parameters
    ----------

    runner : str
        filepath to system runner json

    foreground : str
        name of component
    """
    _, new_path, foreground_federates = remove_from_json(runner, foreground)
    assert len(foreground_federates) == 1
    foreground_fed = foreground_federates[0]
    directory = os.path.join(os.path.dirname(new_path), foreground_fed.directory)
    click.echo("Removing")
    click.echo(f"""
Name      : {foreground_fed.name}
Directory : {directory}
Command   : {foreground_fed.exec}
    """)

    click.echo("Starting system (note you may have to kill manually)")
    helics_sim = subprocess.Popen(["helics", "run", f"--path={new_path}"])
    click.echo(f"Running component {foreground_fed.name} in foreground")
    _ = subprocess.run(foreground_fed.exec.split(), cwd=directory)
    helics_sim.wait()


if __name__ == "__main__":
    cli()



#!/usr/bin/env python3
from oedisi.componentframework.basic_component import component_from_json
from oedisi.componentframework.system_configuration import (
    generate_runner_config,
    WiringDiagram,
)
from oedisi.componentframework.mock_component import MockComponent


def bad_type_checker(type, x):
    return True


TestComponent = component_from_json(
    "test_component_type/component_definition.json", bad_type_checker
)
component_types = {"TestComponent": TestComponent, "MockComponent": MockComponent}

wiring_diagram = WiringDiagram.parse_file("test_basic_system.json")
runner_config = generate_runner_config(wiring_diagram, component_types)

with open("test_system_runner.json", "w") as f:
    f.write(runner_config.json(indent=2))



import os
import json
import pytest
import importlib
from pathlib import Path

DATADIR = Path(__file__).parent / "data"

INPUT_FILES = [p for p in DATADIR.glob("*.json")]

@pytest.mark.parametrize("filepath", INPUT_FILES)
def test_all_formats(filepath):
    name = filepath.name.split('.')[0]
    with open(filepath) as f_in:
        json_data = json.load(f_in)
        class_name = getattr(importlib.import_module("oedisi.types.data_types"),name)
        class_name.parse_obj(json_data)



# -*- coding: utf-8 -*-
"""
"""

import helics as h
import logging
import json
import os


logger = logging.getLogger(__name__)
logger.addHandler(logging.StreamHandler())
logger.setLevel(logging.DEBUG)


def destroy_federate(fed):
    _ = h.helicsFederateDisconnect(fed)
    h.helicsFederateFree(fed)
    h.helicsCloseLibrary()
    logger.info("Federate finalized")


class TestFederate:
    def __init__(self):
        logger.info(f"Current Working Directory: {os.path.abspath(os.curdir)}")
        with open("static_inputs.json") as f:
            self.parameters = json.load(f)

        fedinfo = h.helicsCreateFederateInfo()
        fedinfo.core_name = self.parameters["name"]
        fedinfo.core_type = h.HELICS_CORE_TYPE_ZMQ
        fedinfo.core_init = "--federates=1"

        self.fed = h.helicsCreateValueFederate(self.parameters["name"], fedinfo)
        logger.info(f"Created federate {self.fed.name}")

        with open("input_mapping.json", "r") as f:
            port_mapping = json.load(f)
            self.subscriptions = {}
            if "test1" in port_mapping:
                self.subscriptions["test1"] = self.fed.register_subscription(
                    port_mapping["test1"]
                )
            logging.debug("Loaded subscription test1 at {port_mapping['test1']}")

        self.publications = {}
        self.publications["test2"] = self.fed.register_publication(
            "test2", h.HELICS_DATA_TYPE_DOUBLE
        )

    def run(self):
        self.fed.enter_executing_mode()
        logger.info("Entered HELICS execution mode")

        total_interval = 100
        update_interval = 1
        grantedtime = 0
        while grantedtime < total_interval:
            requested_time = grantedtime + update_interval
            logger.debug(f"Requesting time {requested_time}")
            grantedtime = self.fed.request_time(requested_time)
            logger.debug(f"Granted time {grantedtime}")

            for name, pub in self.fed.publications.items():
                value = 3.14159256535
                logger.info(f"Sending {value} to {name}")
                pub.publish(value)

            for name, sub in self.subscriptions.items():
                if sub.is_updated():
                    logger.info(
                        f"From subscription {name}: {sub.bytes} of type {sub.type}"
                    )

        destroy_federate(self.fed)


if __name__ == "__main__":
    fed = TestFederate()
    fed.run()



"""
Generate basic component from description JSON
"""

import json
import os
from shutil import copytree
from . import system_configuration
from .system_configuration import AnnotatedType
from pydantic import BaseModel
from typing import List, Any, Dict


class ComponentDescription(BaseModel):
    """Component description for simple ComponentType

    Parameters
    ----------
    directory :
        where code is stored relative to where this is run
    execute_function :
        command to execute component
    static_inputs :
        List of types for the parameter
    dynamic_inputs :
        List of input types. Typically subscriptions.
    dynamic_outputs :
        List of output types. Typically publications.
    """
    directory: str
    execute_function: str
    static_inputs: List[AnnotatedType]
    dynamic_inputs: List[AnnotatedType]
    dynamic_outputs: List[AnnotatedType]


def types_to_dict(types: List[AnnotatedType]):
    return {t.port_name: t for t in types}


def component_from_json(f, type_checker):
    comp_desc = ComponentDescription.parse_file(f)
    return basic_component(comp_desc, type_checker)


def basic_component(comp_desc: ComponentDescription, type_checker):
    """Uses data in component_definition to create a new component type

    Parameters
    ----------
    comp_desc : ComponentDescription
         Simplified component representation usually from a JSON file
    type_checker: function taking the type and value and returning a boolean

    Returns
    -------
    BasicComponent(system_configuration.ComponentType) :
         ComponentType from the description
    """

    class BasicComponent(system_configuration.ComponentType):
        _origin_directory = comp_desc.directory
        _execute_function = comp_desc.execute_function
        _dynamic_inputs = types_to_dict(comp_desc.dynamic_inputs)
        _dynamic_outputs = types_to_dict(comp_desc.dynamic_outputs)
        _static_inputs = types_to_dict(comp_desc.static_inputs)

        def __init__(self, name, parameters: Dict[str, Any], directory: str):
            self._name = name
            self._directory = directory
            self._parameters = parameters
            self.check_parameters(parameters)
            self.copy_code_into_directory()
            self.generate_parameter_config()

        def check_parameters(self, parameters):
            for parameter_type in self._static_inputs.values():
                if parameter_type.port_name not in parameters:
                    return False
                if not type_checker(
                        parameter_type.type,
                        parameters[parameter_type.port_name]
                ):
                    return False
            return True

        def copy_code_into_directory(self):
            copytree(self._origin_directory, self._directory, dirs_exist_ok=True)

        def generate_parameter_config(self):
            self._parameters["name"] = self._name
            with open(os.path.join(self._directory, "static_inputs.json"), "w") as f:
                json.dump(self._parameters, f)

        def generate_input_mapping(self, links):
            with open(os.path.join(self._directory, "input_mapping.json"), "w") as f:
                json.dump(links, f)

        @property
        def dynamic_inputs(self):
            return self._dynamic_inputs

        @property
        def dynamic_outputs(self):
            return self._dynamic_outputs

        @property
        def execute_function(self):
            return self._execute_function

    return BasicComponent



"""
This module defines common types and methods for configuration.

The final method `generate_runner_config` brings together
a `WiringDiagram` and a dictionary of `ComponentTypes` with
optional comparison using a compatability checking function.

The `ComponentType` defines the common interface that component
configuration must implement. By default, this can be instantiated
using `basic_component.py`.

The `WiringDiagram` configuration contains a list of components
and the links between them.
"""

from collections import defaultdict
from typing import List, Dict, Type, Any
import os
import logging
import shutil
import psutil
from abc import ABC, abstractmethod, abstractproperty

from pydantic import BaseModel, validator
from typing import List, Optional, Any, Dict


class AnnotatedType(BaseModel):
    "Class for representing the types of components and their interfaces"
    type: str
    description: Optional[str]
    unit: Optional[str]
    port_id: Optional[str]

    @property
    def port_name(self):
        if self.port_id is None:
            return self.type
        return self.port_id


class ComponentType(ABC):
    """Abstract type for component configuration.

    The components define the main restrictions on how components
    can be configured. In the simplest case, the `basic_copmonent` function
    constructs a type from a `ComponentDescription` which just write files.
    There are no restrictions, so for example, one possibility is for the component type
    to interact with a web service.

    First, the class is initialized for each component using the
    name, parameters, and target directory. The federate name
    in HELICS should be the name initialized here.

    Then the `dynamic_inputs` and `dynamic_outputs` are used to check types and verify
    the links used between components. These can depend on the initialziation parameters.
    The `dynamic_outputs` should be initialized under the prefix `name/`.

    Next, `generate_input_mapping` is then called with a mapping
    of the variable names to the HELICS subscription keys. The individual
    federate should then use these names to subscribe at the right location.
    This can also be used for endpoint targets less often.

    Finally, the execute_function property defines the command
    to run the component.
    """
    @abstractmethod
    def generate_input_mapping(self, links: Dict[str, str]):
        pass

    @abstractproperty
    def execute_function(self):
        pass

    @abstractproperty
    def dynamic_inputs(self):
        pass

    @abstractproperty
    def dynamic_outputs(self):
        pass


class Link(BaseModel):
    source: str
    source_port: str
    target: str
    target_port: str


class Port(BaseModel):
    name: str
    port_name: str

    def connect(self, port: 'Port'):
        return Link(
            source=self.name,
            source_port=self.port_name,
            target=port.name,
            target_port=port.port_name
        )

class Component(BaseModel):
    """Component type used in WiringDiagram, includes name,
    component type, and initial parameters"""
    name: str
    type: str
    parameters: Dict[str, Any]

    def port(self, port_name: str):
        return Port(name=self.name, port_name=port_name)


class WiringDiagram(BaseModel):
    "Cosimulation configuration. This may end up wrapped in another interface"
    name: str
    components: List[Component]
    links: List[Link]

    def clean_model(self, target_directory = '.'):
        for component in self.components:
            to_delete = os.path.join(target_directory, component.name)
            log_file = os.path.join(target_directory, component.name+'.log')
            if not os.path.exists(to_delete):
                logging.warning(f"The directory for {component.name} at {to_delete} does not exist")
            else:
                shutil.rmtree(to_delete)
            if not os.path.exists(log_file):
                logging.warning(f"The directory for {component.name} at {log_file} does not exist")
            else:
                os.remove(log_file)


        # TODO: Check for any processes using the HELICS port and kill them too
        for proc in psutil.process_iter():
            if proc.name() == "helics_broker":
                proc.kill()



    @validator("components")
    def check_component_names(cls, components):
        "Check that the components all have unique names"
        names = set(map(lambda c: c.name, components))
        assert len(names) == len(components)
        return components

    @validator("links")
    def check_link_names(cls, links, values, **kwargs):
        if "components" in values:
            components = values["components"]
            names = set(map(lambda c: c.name, components))
            for link in links:
                assert link.source in names and link.target in names
        return links

    def add_component(self, c: Component):
        self.components.append(c)

    def add_link(self, l: Link):
        self.links.append(l)

    @classmethod
    def empty(cls, name="unnamed"):
        return cls(name=name, components=[], links=[])


class Federate(BaseModel):
    "Federate configuration for HELICS CLI"
    directory: str
    hostname: str = "localhost"
    name: str
    exec: str


def initialize_federates(
    wiring_diagram: WiringDiagram,
    component_types: Dict[str, Type[ComponentType]],
    compatability_checker,
    target_directory="."
) -> List[Federate]:
    "Initialize all the federates"
    components = {}
    link_map = get_link_map(wiring_diagram)
    for component in wiring_diagram.components:
        directory = os.path.join(target_directory, component.name)
        if not os.path.exists(directory):
            os.makedirs(directory)
        component_type = component_types[component.type]
        initialized_component = component_type(
            component.name, component.parameters, directory
        )
        components[component.name] = initialized_component

    for l in wiring_diagram.links:
        source_types = components[l.source].dynamic_outputs
        target_types = components[l.target].dynamic_inputs
        assert (
            l.source_port in source_types
        ), f"{l.source} does not have {l.source_port}"
        assert (
            l.target_port in target_types
        ), f"{l.target} does not have dynamic input {l.target_port}"
        source_type = source_types[l.source_port]
        target_type = target_types[l.target_port]
        assert compatability_checker(
            source_type, target_type
        ), f"{source_type} is not compatible with {target_type}"

    federates = []
    for name, component in components.items():
        links = link_map[name]
        component.generate_input_mapping(
            {l.target_port: f"{l.source}/{l.source_port}" for l in links}
        )
        federates.append(
            Federate(directory=name, name=name, exec=component.execute_function)
        )

    return federates


def get_link_map(wiring_diagram: WiringDiagram):
    link_map = defaultdict(list)
    for link in wiring_diagram.links:
        link_map[link.target].append(link)
    return link_map


class RunnerConfig(BaseModel):
    """HELICS running config for the full simulation

    Examples
    --------
    Save to JSON
    
    >>> with open(filename, "w") as f:
    ...    f.write(config.json())

    Run Simulation
    
    `$ helics run --path=filename`
    """
    name: str
    federates: List[Federate]


def bad_compatability_checker(type1, type2):
    "Basic compatability checker that says all types are compatible."
    return True


def generate_runner_config(
    wiring_diagram: WiringDiagram,
    component_types: Dict[str, Type[ComponentType]],
    compatibility_checker=bad_compatability_checker,
    target_directory="."
):
    """Brings together a `WiringDiagram` and a dictionary of `ComponentTypes`
    to create a helics run configuration.

    Parameters
    ----------
    wiring_diagram : WiringDiagram
        Configuration describing components, parameters, and links between them
    component_types : Dict[str, Type[ComponentType]]
        Dictionary for the wiring diagram component types
        to Python component types
    compatibility_checker: function of two types to the booleans
        Each link uses the compatability_checker to ensure the link
        is between compatible types.

    Returns
    -------
    RunnerConfig
        Configuration which can be used to run the cosimulation
    """
    federates = initialize_federates(
        wiring_diagram, component_types, compatibility_checker, target_directory
    )
    broker_federate = Federate(
        directory=".",
        name="broker",
        exec=f"helics_broker -f {len(federates)} --loglevel=warning",
    )
    return RunnerConfig(
        name=wiring_diagram.name, federates=(federates + [broker_federate])
    )



from oedisi.componentframework.mock_component import MockComponent
import os

if not os.path.exists("testC"):
    os.mkdir("testC")

c = MockComponent(
    "federate_name",
    {"outputs": {"pi": "double"}, "inputs": {"possible_input": "double"}},
    "testC",
)

c.generate_input_mapping({})
assert c.execute_function.endswith("componentframework/mock_component.sh")



import helics as h
from pydantic import BaseModel

class TimeData(BaseModel):
    "Time data for a federate"
    name: str
    granted_time: float
    send_time: float


def pprint_time_data(time_data):
    "A table would be better somehow, but which should be the columns"
    print(f"""
    Name         : {time_data.name}
    Granted Time : {time_data.granted_time}
    Send Time    : {time_data.send_time}
    """)


def parse_time_data(response):
    time_data = []
    for core in response["cores"]:
        for fed in core["federates"]:
            time_data.append(TimeData(
                name = fed["attributes"]["name"],
                granted_time = fed["granted_time"],
                send_time = fed["send_time"]
            ))

    return time_data

def get_time_data(broker):
    # Use global time debugging?
    return parse_time_data(
        broker.query("broker", "global_time")
    )



"""
MockComponent and MockFederate allow you to instantiate a mock component
with a specified set of inputs and outputs. The parameters dictionary
should contain a list under "inputs" and "outputs". During implementation,
the value pi is passed around every second until t=100.

MockComponent defines the ComponentType for a simple testing component.

MockFederate defines the corresponding implementation.
"""

import helics as h
import logging
from typing import Dict
import json
import os
from . import system_configuration
from .system_configuration import AnnotatedType


logger = logging.getLogger(__name__)
logger.addHandler(logging.StreamHandler())
logger.setLevel(logging.DEBUG)


class MockComponent(system_configuration.ComponentType):
    def __init__(self, name, parameters: Dict[str, Dict[str, str]], directory: str):
        self._name = name
        self._directory = directory
        self._execute_function = os.path.join(
            os.path.dirname(os.path.abspath(__file__)), "mock_component.sh"
        )
        self.process_parameters(parameters)

    def process_parameters(self, parameters):
        self._dynamic_inputs = {
            name: AnnotatedType(type="", port_id=name) for name in parameters["inputs"]
        }
        self._dynamic_outputs = {
            name: AnnotatedType(type=type, port_id=name)
            for name, type in parameters["outputs"].items()
        }
        self.generate_helics_config(parameters["outputs"])

    def generate_helics_config(self, outputs):
        helics_config = {
            "name": self._name,
            "core_type": "zmq",
            "period": 1,
            "log_level": "warning",
            "terminate_on_error": True,
            "publications": [
                {"key": key, "type": value} for key, value in outputs.items()
            ],
        }

        with open(os.path.join(self._directory, "helics_config.json"), "w") as f:
            json.dump(helics_config, f)

    def generate_input_mapping(self, links):
        with open(os.path.join(self._directory, "input_mapping.json"), "w") as f:
            json.dump(links, f)

    @property
    def dynamic_inputs(self):
        return self._dynamic_inputs

    @property
    def dynamic_outputs(self):
        return self._dynamic_outputs

    @property
    def execute_function(self):
        return self._execute_function


def get_default_value(date_type: h.HelicsDataType):
    return 3.1415926536


def destroy_federate(fed):
    _ = h.helicsFederateDisconnect(fed)
    h.helicsFederateFree(fed)
    h.helicsCloseLibrary()
    logger.info("Federate finalized")


class MockFederate:
    def __init__(self):
        logger.info(f"Current Working Directory: {os.path.abspath(os.curdir)}")
        self.fed = h.helicsCreateValueFederateFromConfig("helics_config.json")
        logger.info(f"Created federate {self.fed.name}")

        with open("input_mapping.json", "r") as f:
            port_mapping = json.load(f)
            self.subscriptions = {}
            for name, key in port_mapping.items():
                self.subscriptions[name] = self.fed.register_subscription(key)
                logging.debug("Loaded subscription {name} at {key}")
            logging.info("Loaded all subscriptions from file")

    def run(self):
        self.fed.enter_executing_mode()
        logger.info("Entered HELICS execution mode")

        total_interval = 100
        update_interval = 1
        grantedtime = 0
        while grantedtime < total_interval:
            requested_time = grantedtime + update_interval
            logger.debug(f"Requesting time {requested_time}")
            grantedtime = self.fed.request_time(requested_time)
            logger.debug(f"Granted time {grantedtime}")

            for name, pub in self.fed.publications.items():
                value = get_default_value(pub.type)
                logger.info(f"Sending {value} to {name}")
                pub.publish(value)

            for name, sub in self.subscriptions.items():
                if sub.is_updated():
                    logger.info(
                        f"From subscription {name}: {sub.bytes} of type {sub.type}"
                    )

        destroy_federate(self.fed)


if __name__ == "__main__":
    fed = MockFederate()
    fed.run()



from __future__ import annotations
import datetime
from enum import Enum
from pydantic import BaseModel, root_validator
from typing import List, Optional, Union, Tuple

### Supporting Functions ###
# TODO: Connect with CIM values


Complex = Tuple[float, float]


class StateArray(BaseModel):
    """
    Extended by classes:
        "SwitchStates",
        "CapacitorStates",
        "RegulatorStates"

    """

    values: List[int]
    ids: List[str]
    time: Optional[datetime.datetime]


class SwitchStates(StateArray):
    pass


class CapacitorStates(StateArray):
    pass


class RegulatorStates(StateArray):
    pass


class CostArray(BaseModel):
    """
    Extended by classes:
        "RealCostFunctions",
        "ReactiveCostFunctions",
        "RealWholesalePrices",
        "ReactiveWholesalePrices",
        "OperationalCosts"

    """

    values: List[List[float]]
    ids: List[str]
    units: str = "$"
    time: Optional[datetime.datetime]


class RealCostFunctions(CostArray):
    pass


class ReactiveCostFunctions(CostArray):
    pass


class RealWholesalePrices(CostArray):
    pass


class ReactiveWholesalePrices(CostArray):
    pass


class OperationalCosts(CostArray):
    pass


class MeasurementArray(BaseModel):
    """
    Extended by classes:
        "BusArray",
        "EquipmentArray",
        "EquipmentNodeArray"
    """

    values: List[float]
    ids: List[str]
    units: str
    accuracy: Optional[List[float]]
    bad_data_threshold: Optional[List[float]]
    time: Optional[datetime.datetime]


class BusArray(MeasurementArray):
    """
    Extended by classes:
        "VoltagesMagnitude",
        "VoltagesAngle",
        "VoltagesReal",
        "VoltagesImaginary"
    """

    pass


class EquipmentArray(MeasurementArray):
    """
    Extended by classes:
        "SolarIrradiances",
        "Temperatures",
        "WindSpeeds",
        "StatesOfCharge",
        "CurrentsMagnitude",
        "CurrentsAngle",
        "CurrentsReal",
        "CurrentsImaginary",
        "ImpedanceMagnitude",
        "ImpedanceAngle",
        "ImpedanceReal",
        "ImpedanceImaginary",
    """

    pass


class EquipmentNodeArray(MeasurementArray):
    """
    Primary key is ids + equipment_ids.

    - ids corresponding node id, so "113.1", "113.2", "113.3"
    - equipment_id corresponds to PVSystem.113

    Extended by classes:
        "PowersMagnitude",
        "PowersAngle",
        "PowersReal",
        "PowersImaginary",

    """

    equipment_ids: List[str]


class VoltagesMagnitude(BusArray):
    units: str = "kV"


class VoltagesAngle(BusArray):
    units: str = "degrees"


class VoltagesReal(BusArray):
    units: str = "kV"


class VoltagesImaginary(BusArray):
    units: str = "kV"


class CurrentsMagnitude(EquipmentArray):
    units: str = "A"


class CurrentsAngle(EquipmentArray):
    units: str = "degrees"


class CurrentsReal(EquipmentArray):
    units: str = "A"


class CurrentsImaginary(EquipmentArray):
    units: str = "A"


class ImpedanceReal(EquipmentArray):
    units: str = "Ohm"


class ImpedanceImaginary(EquipmentArray):
    units: str = "Ohm"


class ImpedanceMagnitude(EquipmentArray):
    units: str = "Ohm"


class ImpedanceAngle(EquipmentArray):
    units: str = "degrees"


class PowersMagnitude(EquipmentNodeArray):
    units: str = "kVA"


class PowersAngle(EquipmentNodeArray):
    units: str = "degrees"


class PowersReal(EquipmentNodeArray):
    units: str = "kW"


class PowersImaginary(EquipmentNodeArray):
    units: str = "kVAR"


class SolarIrradiances(EquipmentArray):
    units: str = "kW/m^2"


class Temperatures(EquipmentArray):
    units: str = "C"


class WindSpeeds(EquipmentArray):
    units: str = "m/s"


class StatesOfCharge(EquipmentArray):
    units: str = "percent"


class Topology(BaseModel):
    admittance: Union[AdmittanceSparse, AdmittanceMatrix]
    injections: Injection
    incidences: Optional[Incidence]
    base_voltage_angles: Optional[VoltagesAngle]
    base_voltage_magnitudes: Optional[VoltagesMagnitude]
    slack_bus: List[str] = []


class Incidence(BaseModel):
    from_equipment: List[str]
    to_equipment: List[str]
    equipment_type: Optional[List[str]]


class AdmittanceSparse(Incidence):
    admittance_list: List[Complex]
    units: str = "S"


class AdmittanceMatrix(BaseModel):
    admittance_matrix: List[List[Complex]]
    ids: List[str]
    units: str = "S"


class Injection(BaseModel):
    # Shouldn't these be equipment arrays?
    current_real: CurrentsReal = {"values": [], "ids": [], "node_ids": []}
    current_imaginary: CurrentsImaginary = {"values": [], "ids": [], "node_ids": []}
    power_real: PowersReal = {"values": [], "ids": [], "node_ids": []}
    power_imaginary: PowersImaginary = {"values": [], "ids": [], "node_ids": []}
    impedance_real: ImpedanceReal = {"values": [], "ids": [], "node_ids": []}
    impedance_imaginary: ImpedanceImaginary = {"values": [], "ids": [], "node_ids": []}


Topology.update_forward_refs()


class Command(BaseModel):
    """JSON Configuration for external object commands.

    obj_name -- name of the object.
    obj_prop -- name of the property.
    obj_val -- val of the property.
    """

    obj_name: str
    obj_property: str
    val: Union[int, float, str, List[int], List[float], List[str]]


class CommandList(BaseModel):
    """List[Command] with JSON parsing."""

    __root__: List[Command]


class ReactivePowerSetting(Enum):
    """Reactive power setting, almost always VARAVAL_WATTS."""

    VARAVAL_WATTS = "VARAVAL_WATTS"
    VARMAX_VARS = "VARMAX_VARS"
    VARMAX_WATTS = "VARMAX_WATTS"


class InverterControlMode(Enum):
    """Inverter control mode."""

    voltvar = "VOLTVAR"
    voltwatt = "VOLTWATT"
    voltvar_voltwatt = "VV_VW"


class VVControl(BaseModel):
    """OpenDSS setting for volt-var control."""

    deltaq_factor: float = -1.0  # -1.0 tells OpenDSS to figure it out
    varchangetolerance: float = 0.025
    voltagechangetolerance: float = 0.0001
    vv_refreactivepower: ReactivePowerSetting = ReactivePowerSetting.VARAVAL_WATTS
    voltage: List[float]  # p.u. in V
    reactive_response: List[float]  # p.u. in VArs


class VWControl(BaseModel):
    """OpenDSS setting for volt-watt control."""

    deltap_factor: float = -1.0  # -1.0 tells OpenDSS to figure it out
    voltage: List[float]  # p.u. in V
    power_response: List[float]  # p.u. in VArs


class InverterControl(BaseModel):
    """InverterControl with volt-var control and/or volt-watt control."""

    pvsystem_list: Optional[List[str]] = None
    vvcontrol: Optional[VVControl] = None
    vwcontrol: Optional[VWControl] = None
    mode: InverterControlMode = InverterControlMode.voltvar

    @root_validator(pre=True)
    def check_mode(cls, values):
        """Make sure that mode reflects vvcontrol and vwcontrol data."""
        if "mode" not in values or (
            values["mode"] == InverterControlMode.voltvar
            or values["mode"] == InverterControlMode.voltvar_voltwatt
        ):
            assert "vvcontrol" in values and values["vvcontrol"] is not None
        if "mode" in values and (
            values["mode"] == InverterControlMode.voltwatt
            or values["mode"] == InverterControlMode.voltvar_voltwatt
        ):
            assert "vwcontrol" in values and values["vwcontrol"] is not None
        return values


class InverterControlList(BaseModel):
    """List[InverterControl] with JSON parsing."""

    __root__: List[InverterControl]



import time

import helics as h


def get_inputs_outputs(graph_dict):
    federate_inputs = {}
    federate_outputs = {}
    federate_id_handle2key = {}

    federate_id2name = {}
    for c in graph_dict["cores"]:
        for fed in c["federates"]:
            name = fed["attributes"]["name"]
            federate_id2name[
                fed["attributes"]["id"]
            ] = name
            pub_keys = []
            for pub in fed.get("publications", []):
                federate_id_handle2key[
                    (pub["federate"], pub["handle"])
                ] = pub["key"]
                pub_keys.append(pub["key"])
            federate_outputs[name] = pub_keys

    for c in graph_dict["cores"]:
        for fed in c["federates"]:
            name = fed["attributes"]["name"]
            sub_keys = []
            for sub in fed.get("inputs", []):
                for source in sub["sources"]:
                    sub_keys.append(federate_id_handle2key[
                        (source["federate"], source["handle"])
                    ])
            federate_inputs[name] = sub_keys
    return federate_inputs, federate_outputs


class TestingBroker:
    def __init__(self, n):
        self.initstring = f"-f {n} --name=mainbroker"

    def run(self):
        time.sleep(2)
        self.broker = h.helicsCreateBroker("zmq", "", self.initstring)
        h.helicsBrokerSetTimeBarrier(self.broker, 0.0)
        print("Initialized broker")
        # Wait until federates have hit the time barrier
        self.wait_until_connected()

        graph_dict = self.broker.query("broker", "data_flow_graph")
        self.broker.disconnect()
        return get_inputs_outputs(graph_dict)

    def wait_until_connected(self):
        print("Waiting for initialization")
        while True:
            time.sleep(2)
            current_state = self.broker.query("broker", "current_state")
            cores = current_state["cores"]
            if (len(cores) == 2 and
                all((core["state"] == "init_requested" for core in cores))):
                return



# -*- coding: utf-8 -*-
"""
setup.py
"""
import os
from codecs import open
from setuptools import setup, find_packages
from setuptools.command.develop import develop
from subprocess import check_call
import shlex
from warnings import warn

here = os.path.abspath(os.path.dirname(__file__))

description = (
    "Orchestration interface for HELICS power simulations"
)

with open("README.md") as f:
    readme = f.read()

setup(
    name="oedisi",
    version="1.0.0rc1",
    description=description,
    long_description=readme,
    long_description_content_type='text/markdown',
    author="Joseph McKinsey",
    author_email="joseph.mckinsey@nrel.gov",
    packages=find_packages(),
    package_dir={"oedisi": "oedisi"},
    include_package_data=True,
    license="BSD 3-Clause",
    zip_safe=True,
    keywords="oedisi gadal helics",
    python_requires=">=3.8",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    # test_suite="tests",
    install_requires=["helics-apps>=3.2.1", "helics>=3.2.1", "pydantic>=1.7,<2", "psutil", "click"],
    extras_require={"test":["pytest"]},
    entry_points = {
        'console_scripts': [
            'oedisi = oedisi.tools:cli'
        ]
    }
    # cmdclass={"develop": PostDevelopCommand},
)



from oedisi.componentframework.system_configuration import (
    generate_runner_config,
    WiringDiagram,
)
from oedisi.componentframework.mock_component import MockComponent
import json

component_types = {"MockComponent": MockComponent}
wiring_diagram = WiringDiagram.parse_file("test_system.json")
runner_config = generate_runner_config(wiring_diagram, component_types, target_directory="build")

with open("build/test_system_runner.json", "w") as f:
    f.write(runner_config.json(indent=2))
